
// Name
//===================================
//
// Sprite
//


// DEPENDENCIES
//===================================

    #include "collision.h"
    #include "sprite.h"
    #include "sprite_graphics.h"
    #include "sprite_movement.h"
    #include "input.h"
    #include "unit.h"


// STATIC PROPERTIES
//===================================

// METHODS
//===================================

    Sprite::Sprite
    (
        Graphics::SpriteSheet texture,
        SpriteGraphics* graphics,
        SpriteMovement* movement,
        int x,
        int y,
        int width,
        int height,
        Type type,
        int texture_x,
        int texture_y
    )
    :
        graphics_ ( graphics ),
        movement_ ( movement ),
        texture_ ( texture ),
        width_ ( width ),
        height_ ( height ),
        physics_bound_box_ ( { Unit::PixelsToSubPixels( x + 1), Unit::PixelsToSubPixels( y + 1), Unit::PixelsToSubPixels( width_ - 2 ), Unit::PixelsToSubPixels( height_ - 1 ) } ),
        type_ ( type ),
        texture_x_ ( texture_x ),
        texture_y_ ( texture_y )
    {
        adjustBoundBoxes();
    };

    Sprite::~Sprite()
    {
        delete movement_;
    };

    void Sprite::update( Input& input )
    {

        if ( movement_ != nullptr )
            movement_->update( *this, input );

        position();
        adjustBoundBoxes();
        boundaries();
        adjustBoundBoxes();

        if ( graphics_ != nullptr )
        {
            graphics_->update( *this );
        }
    };

    void Sprite::render( Graphics& graphics )
    {
        if ( graphics_ != nullptr )
        {
            graphics_->render( graphics_bound_box_, graphics );
        }
        else
        {
            SDL_Rect src = { texture_x_, texture_y_, graphics_bound_box_.w, graphics_bound_box_.h };
            graphics.renderObject( texture_, src, graphics_bound_box_ );
        }
    };

    void Sprite::moveLeft()
    {
        acceleration_x_ = -start_speed_;
        direction_x_ = Direction::Horizontal::LEFT;
    };

    void Sprite::moveRight()
    {
        acceleration_x_ = start_speed_;
        direction_x_ = Direction::Horizontal::RIGHT;
    };

    void Sprite::stop()
    {
        acceleration_x_ = 0;
    };

    void Sprite::stopDucking()
    {
        acceleration_x_ = 0;
    };

    void Sprite::position()
    {
        x_prev_ = physics_bound_box_.x;
        y_prev_ = physics_bound_box_.y;

        if ( isJumping() )
        {
            acceleration_x_ *= .75;
        }

        if ( !onGround() )
            top_speed_ = TOP_SPEED_WALK * .95;

        if ( isDucking() )
        {
            acceleration_x_ /= 3;
            jump_top_speed_ = JUMP_DUCK_TOP_SPEED;
        }

        vx_ += acceleration_x_;

        if ( acceleration_x_ == 0 )
            vx_ /= 1.1;

        if ( vx_ > top_speed_ )
            vx_ = top_speed_;
        if ( vx_ < -top_speed_ )
            vx_ = -top_speed_;

        physics_bound_box_.x += vx_;

        if ( acceleration_x_ != 0 )
            is_moving_ = true;
        else
            is_moving_ = false;

        top_speed_ = TOP_SPEED_WALK;
        start_speed_ = START_SPEED_WALK;

        is_running_ = false;


        // VERTICAL MOVEMENT

        if ( abs( top_speed_ ) > TOP_SPEED_WALK )
            jump_top_speed_ *= 1.5;

        if ( !onGround() && !isJumping() && !onLadder() )
        {
            acceleration_y_ = gravity_start_speed_;
        }
        /*
        else if ( !isJumping() )
        {
            if ( vy_ < 0 )
                vy_ /= 1.5;
        }
        */

        vy_ += acceleration_y_;

        if ( vy_ > gravity_top_speed_ )
            vy_ = gravity_top_speed_;

        if ( vy_ < -jump_top_speed_ )
        {
            vy_ = -jump_top_speed_;
            jump_start_ = false;
        }

        physics_bound_box_.y += vy_;

        gravity_start_speed_ = GRAVITY_START_SPEED;
        gravity_top_speed_ = GRAVITY_TOP_SPEED;
        jump_top_speed_ = JUMP_TOP_SPEED;

        acceleration_y_ = 0;

        if ( !isJumping() )
            jump_end_ = true;

        is_jumping_ = false;
        on_ground_prev_ = on_ground_;
        on_ground_ = false;

        if ( onLadder() )
        {
            vy_ = 0;
        }
        //on_ladder_prev_ = on_ladder_;
        //on_ladder_ = false;

        //touching_ladder_prev_ = touching_ladder_;
        //touching_ladder_ = false;
    };

    void Sprite::boundaries()
    {
        if ( physics_bound_box_.x < 0 )
        {
            //physics_bound_box_.x = 0;
            collideStopXLeft( abs ( 0 - physics_bound_box_.x ) );
        }
        if ( physics_bound_box_.x + physics_bound_box_.w > Unit::PixelsToSubPixels( Unit::WINDOW_WIDTH_PIXELS ) )
        {
            //physics_bound_box_.x = Unit::PixelsToSubPixels( Unit::WINDOW_WIDTH_PIXELS - ( halfWidth() - 2) );
            collideStopXRight( ( physics_bound_box_.x + physics_bound_box_.w ) - Unit::PixelsToSubPixels( Unit::WINDOW_WIDTH_PIXELS ) );
        }

        /*
        if ( physics_bound_box_.y + physics_bound_box_.h > Unit::PixelsToSubPixels( Unit::WINDOW_HEIGHT_PIXELS ) )
        {
            //physics_bound_box_.y = Unit::PixelsToSubPixels( Unit::WINDOW_HEIGHT_PIXELS ) - physics_bound_box_.h;
            collideStopYBottom( 0 );
        }*/
    };

    void Sprite::collideStopXLeft( int overlap )
    {
        stop();
        if ( vx_ < 0 )
        {
            physics_bound_box_.x -= vx_;
            vx_ = -(vx_ * BOUNCE);
            //physics_bound_box_.x += overlap;
        }
    };

    void Sprite::collideStopXRight( int overlap )
    {
        stop();
        if ( vx_ > 0 )
        {
            physics_bound_box_.x -= vx_;
            vx_ = -(vx_ * BOUNCE);
            //physics_bound_box_.x -= overlap;
        }
    };

    void Sprite::collideStopYBottom( int overlap )
    {
        if ( vy_ > 0 )
        {
            acceleration_y_ = 0;
            //physics_bound_box_.y -= vy_;

            if ( !onGround() )
            {
                //physics_bound_box_.y += 1000;
            }

            //vy_ = 0;

            if ( !onGround() )
            {
                physics_bound_box_.y -= overlap;
            }

            on_ground_ = true;
            //is_jumping_ = false;
        }
    };

    void Sprite::collideStopYTop( int overlap )
    {
        acceleration_y_ = 0;
        physics_bound_box_.y += vy_;
        vy_ = 0;
        physics_bound_box_.y += overlap;
        jump_end_ = true;
    };

    bool Sprite::onGround() const
    {
        return on_ground_;
    };

    bool Sprite::isJumping() const
    {
        return is_jumping_;
    };

    void Sprite::jump()
    {
        if ( !onGround() || !jump_lock_ )
        {
        if ( onGround() && !jump_lock_ )
        {
            jump_start_ = true;
            jump_end_ = false;
            vy_ = 0;
            on_ground_ = false;
        }

        if ( jump_start_ && !jump_end_ )
        {
            is_jumping_ = true;
        }

        if ( isJumping() )
        {
            acceleration_y_ = -JUMP_START_SPEED;
        }
        }

    };

    void Sprite::slowFall()
    {
        gravity_start_speed_ = GRAVITY_START_SPEED * .75;
        gravity_top_speed_ = GRAVITY_TOP_SPEED * .85;
    };

    bool Sprite::isRunning() const
    {
        return is_running_;
    };

    void Sprite::run()
    {
        is_running_ = true;
        top_speed_ = TOP_SPEED_RUN;
        start_speed_ = START_SPEED_RUN;
    };


    void Sprite::adjustBoundBoxes()
    {
        /*
        physics_bound_box_ =
        {
            xSubPixels() - Unit::PixelsToSubPixels( halfWidth() - 1 ),
            ySubPixels() - Unit::PixelsToSubPixels( height_ - 2 ),
            Unit::PixelsToSubPixels( width_ - 2 ),
            Unit::PixelsToSubPixels( height_ - 2 )
        };
        */

        graphics_bound_box_ = { xPixels() - 1, yPixels() - 1, width_, heightPixels() + 1 };
    };


    int Sprite::heightPixels() const
    {
        return Unit::SubPixelsToPixels( physics_bound_box_.h );
    };

    int Sprite::xPixels() const
    {
        return Unit::SubPixelsToPixels( physics_bound_box_.x );
    };
    int Sprite::yPixels() const
    {
        return Unit::SubPixelsToPixels( physics_bound_box_.y );
    };
    int Sprite::xSubPixels() const
    {
        return physics_bound_box_.x;
    };
    int Sprite::ySubPixels() const
    {
        return physics_bound_box_.y;
    };

    int Sprite::hitboxHalfWidth() const
    {
        return round( physics_bound_box_.w / 2 );
    };
    int Sprite::hitboxHalfHeight() const
    {
        return round( physics_bound_box_.h / 2 );
    };

    int Sprite::hitboxCenterX() const
    {
        return xSubPixels() + hitboxHalfWidth();
    };

    int Sprite::hitboxCenterY() const
    {
        return ySubPixels() + hitboxHalfHeight();
    };

    void Sprite::collisionDetect( Sprite& other )
    {
        if ( movement_ != nullptr )
            movement_->collision( collide( other ), *this, other );

        adjustBoundBoxes();
    };

    Collision Sprite::collide( Sprite& other ) const
    {
        int overlap_x = 0;
        int overlap_y = 0;
        int overlap_x_left   = 0;
        int overlap_x_right  = 0;
        int overlap_y_top    = 0;
        int overlap_y_bottom = 0;
        Direction::Simple direction = Direction::Simple::__NULL;
        Direction::Horizontal direction_x = Direction::Horizontal::__NULL;
        Direction::Vertical direction_y = Direction::Vertical::__NULL;

        /*
        double dx = hitboxCenterX() - other.hitboxCenterX();
        double dy = hitboxCenterY() - other.hitboxCenterY();

        double hx = ( other.hitboxHalfWidth() ) + ( hitboxHalfWidth() );
        double hy = ( other.hitboxHalfHeight() ) + ( hitboxHalfHeight() );

        bool collision = abs( dx ) < hx && abs( dy ) < hy;

        if ( collision )
        {
            overlap_x = hx - abs( dx );
            overlap_y = hy - abs( dy );

            ( dy > 0 ) ? direction_y = Direction::Vertical::UP : direction_y = Direction::Vertical::DOWN;
            ( dx > 0 ) ? direction_x = Direction::Horizontal::LEFT : direction_x = Direction::Horizontal::RIGHT;

            if ( overlap_x > overlap_y )
            {
                direction = Direction::verticalToSimple( direction_y );
            }
            else
            {
                direction = Direction::horizontalToSimple( direction_x );
            }
        }*/

        int left_bottom_point_x = physics_bound_box_.x + 2000;
        int right_bottom_point_x = physics_bound_box_.x + physics_bound_box_.w - 2000;
        int center_bottom_point_x = physics_bound_box_.x + hitboxHalfHeight();
        int bottom_points_y = physics_bound_box_.y + physics_bound_box_.h;

        int target_downward_left_bound_x = other.physics_bound_box_.x;
        int target_downward_right_bound_x = other.physics_bound_box_.x + other.physics_bound_box_.w;
        int target_downward_top_bound_y = other.physics_bound_box_.y;
        int target_downward_bottom_bound_y = other.physics_bound_box_.y + other.physics_bound_box_.h;

        if
        (
            ( left_bottom_point_x > target_downward_left_bound_x && left_bottom_point_x < target_downward_right_bound_x ) ||
            ( right_bottom_point_x > target_downward_left_bound_x && right_bottom_point_x < target_downward_right_bound_x ) ||
            ( center_bottom_point_x > target_downward_left_bound_x && center_bottom_point_x < target_downward_right_bound_x )
        )
        {
            if
            (
                bottom_points_y > target_downward_top_bound_y &&
                bottom_points_y < target_downward_bottom_bound_y
            )
            {
                direction_y = Direction::Vertical::DOWN;
                overlap_y_bottom = bottom_points_y - target_downward_top_bound_y;
            }
        }

        int top_points_y = physics_bound_box_.y;

        int target_upward_top_bound_y = other.physics_bound_box_.y + other.hitboxHalfHeight();
        int target_upward_bottom_bound_y = other.physics_bound_box_.y + other.physics_bound_box_.h;

        if
        (
            ( left_bottom_point_x > target_downward_left_bound_x && left_bottom_point_x < target_downward_right_bound_x ) ||
            ( right_bottom_point_x > target_downward_left_bound_x && right_bottom_point_x < target_downward_right_bound_x ) ||
            ( center_bottom_point_x > target_downward_left_bound_x && center_bottom_point_x < target_downward_right_bound_x )
        )
        {
            if
            (
                top_points_y < target_upward_bottom_bound_y &&
                top_points_y > target_upward_top_bound_y
            )
            {
                direction_y = Direction::Vertical::UP;
                overlap_y_top = target_upward_bottom_bound_y - top_points_y;
            }
        }


        int left_point_x = physics_bound_box_.x;
        int left_point_y_top = physics_bound_box_.y;
        int left_point_y_center = physics_bound_box_.y + hitboxHalfHeight();
        int left_point_y_bottom = physics_bound_box_.y + physics_bound_box_.h - 8000;

        int left_target_bound_x_left = other.physics_bound_box_.x + other.hitboxHalfWidth();
        int left_target_bound_x_right = other.physics_bound_box_.x + other.physics_bound_box_.w;
        int left_target_bound_y_top = other.physics_bound_box_.y;
        int left_target_bound_y_bottom = other.physics_bound_box_.y + other.physics_bound_box_.h;

        if
        (
            ( left_point_y_top > left_target_bound_y_top && left_point_y_top < left_target_bound_y_bottom ) ||
            ( left_point_y_bottom > left_target_bound_y_top && left_point_y_bottom < left_target_bound_y_bottom ) ||
            ( left_point_y_center > left_target_bound_y_top && left_point_y_center < left_target_bound_y_bottom )
        )
        {
            if
            (
                left_point_x > left_target_bound_x_left &&
                left_point_x < left_target_bound_x_right
            )
            {
                direction_x = Direction::Horizontal::LEFT;
                overlap_x_left = left_target_bound_x_right - left_point_x;
            }
        }


        int right_point_x = physics_bound_box_.x + physics_bound_box_.w;

        int right_target_bound_x_left = other.physics_bound_box_.x;
        int right_target_bound_x_right = other.physics_bound_box_.x + other.hitboxHalfWidth();

        if
        (
            ( left_point_y_top > left_target_bound_y_top && left_point_y_top < left_target_bound_y_bottom ) ||
            ( left_point_y_bottom > left_target_bound_y_top && left_point_y_bottom < left_target_bound_y_bottom ) ||
            ( left_point_y_center > left_target_bound_y_top && left_point_y_center < left_target_bound_y_bottom )
        )
        {
            if
            (
                right_point_x > right_target_bound_x_left &&
                right_point_x < right_target_bound_x_right
            )
            {
                direction_x = Direction::Horizontal::RIGHT;
                overlap_x_right = right_point_x - right_target_bound_x_left;
            }
        }


        return Collision( direction, overlap_x, overlap_y, direction_x, direction_y, overlap_x_left, overlap_x_right, overlap_y_top, overlap_y_bottom );
    };

    bool Sprite::canJump() const
    {
        return can_jump_;
    };

    void Sprite::debugResetPos()
    {
        physics_bound_box_.x = 96000;
        physics_bound_box_.y = 64000;
    };

    bool Sprite::isDucking() const
    {
        return is_ducking_;
    };

    void Sprite::duck()
    {
        if ( isDucking() || onGround() )
        {
            if ( !isDucking() )
                physics_bound_box_.y += 8000;

            is_ducking_ = true;

            physics_bound_box_.h = Unit::PixelsToSubPixels( height_ - 7 );
        }
    };

    void Sprite::unduck()
    {
        if ( isDucking() )
            physics_bound_box_.y -= 8000;

        is_ducking_ = false;
        physics_bound_box_.h = Unit::PixelsToSubPixels( height_ - 1 );
    };


    Direction::Horizontal Sprite::directionX() const
    {
        return direction_x_;
    };

    bool Sprite::onGroundPrev() const
    {
        return on_ground_prev_;
    };

    bool Sprite::isMoving() const
    {
        return is_moving_;
    };

    void Sprite::stopY()
    {
        acceleration_y_ = 0;
        vy_ = 0;
    };

    Sprite::Type Sprite::type() const
    {
        return type_;
    };


    bool Sprite::onLadder() const
    {
        return on_ladder_;
    };
    bool Sprite::onLadderPrev() const
    {
        return on_ladder_prev_;
    };
    bool Sprite::touchingLadder() const
    {
        return touching_ladder_;
    };
    bool Sprite::touchingLadderPrev() const
    {
        return touching_ladder_prev_;
    };

    void Sprite::touchLadder()
    {
        touching_ladder_ = true;
    };
    void Sprite::grabLadder()
    {
        on_ladder_ = true;
    };
    void Sprite::releaseLadder()
    {
        on_ladder_ = false;
    };
    void Sprite::moveUp()
    {
        if ( onLadder() )
        {
            acceleration_y_ = -1000;
        }
    };
    void Sprite::moveDown()
    {
        if ( onLadder() )
        {
            acceleration_y_ = +1000;
        }
    };


    int Sprite::xPrevSubPixels() const
    {
        return x_prev_;
    };
    int Sprite::yPrevSubPixels() const
    {
        return y_prev_;
    };
    int Sprite::xPrevPixels() const
    {
        return Unit::SubPixelsToPixels( x_prev_ );
    };
    int Sprite::yPrevPixels() const
    {
        return Unit::SubPixelsToPixels( y_prev_ );
    };
